extends Node

const SAVE_PATH = "user://savegame_slot_1.tres"

var state: GameState
var random_gen := RandomNumberGenerator.new()

func _init():
	load_state()

func start_new_run(player_states: Array[PlayerState]):
	random_gen.randomize()
	randomize()
	
	state.level = 1
	
	state.run_souls = 0
	state.coins = 0
	state.luck_factor = 1.0
	state.nb_bombs = 0
	state.nb_keys = 0
	state.used_bombs = 0
	state.used_keys = 0
	state.heart_bank_machine_broken = false
	state.coin_bank_machine_broken = false
	state.gods_offended = false
	
	state.players = player_states
	state.previous_rooms = []
	state.nb_horsemen_killed = 0
	
	var initial_players: Array[PlayerState] = []
	for player in player_states:
		initial_players.append(player.duplicate())
	state.initial_players = initial_players
	
	state.room_type = Rooms.RoomType.start
	generate_doors_for_current_room()
	
	persist_state()

func set_next_level(room_type: Rooms.RoomType):
	state.level += 1
	state.room_type = room_type
	if state.get_room_number() == 1:
		state.previous_rooms.clear()
	else:
		state.previous_rooms.append(room_type)
	generate_doors_for_current_room()
	
	persist_state()

func set_specific_level(level_number: int):
	state.level = level_number
	if state.get_room_number() == 1:
		state.room_type = Rooms.RoomType.start
	elif state.get_room_number() == 10:
		state.room_type = Rooms.RoomType.boss
	else:
		state.room_type = Rooms.RoomType.shop
	generate_doors_for_current_room()
	
	persist_state()

func quit_run():
	state.level = 0 # we only need to mark level at 0 (invalid value)
	persist_state()

func persist_state(save_seed := true):
	if save_seed:
		state.run_seed = random_gen.state
	ResourceSaver.save(state, SAVE_PATH)

func load_state():
	if not ResourceLoader.exists(SAVE_PATH):
		state = GameState.new()
		return
	else:
		# We need to ignore the cache, otherwise we will keep the altered state which is already in memory
		state = ResourceLoader.load(SAVE_PATH, "", ResourceLoader.CACHE_MODE_IGNORE)
		random_gen.state = state.run_seed

# More you have luck, more the result will be high
func luck_based_randi_range(from: int, to: int) -> int:
	var initial_result := random_gen.randi_range(from, to)
	
	return min(initial_result * state.luck_factor, to)

func generate_doors_for_current_room(force_new_ones := false):
	if state.is_first_final_boss_room():
		state.left_door_type = Rooms.RoomType.none
		state.left_door_protection = Rooms.DoorProtection.none
		state.right_door_type = Rooms.RoomType.start
		state.right_door_protection = Rooms.DoorProtection.boss_key
		return
	if state.is_second_final_boss_room():
		# No more doors (at least for now)
		state.left_door_type = Rooms.RoomType.none
		state.left_door_protection = Rooms.DoorProtection.none
		state.right_door_type = Rooms.RoomType.none
		state.right_door_protection = Rooms.DoorProtection.none
		return
	if state.is_boss_room():
		state.left_door_type = Rooms.RoomType.none
		state.left_door_protection = Rooms.DoorProtection.none
		state.right_door_type = Rooms.RoomType.start
		state.right_door_protection = Rooms.DoorProtection.none
		return
	if state.is_before_boss_room():
		state.left_door_type = Rooms.RoomType.none
		state.left_door_protection = Rooms.DoorProtection.none
		state.right_door_type = Rooms.RoomType.boss
		state.right_door_protection = Rooms.DoorProtection.none
		return
	
	var possible_rooms := Rooms.RoomType.values()
	possible_rooms.erase(Rooms.RoomType.none)
	possible_rooms.erase(Rooms.RoomType.start)
	possible_rooms.erase(Rooms.RoomType.boss)
	possible_rooms.erase(Rooms.RoomType.newspaper)
	
	if force_new_ones:
		possible_rooms.erase(state.left_door_type)
		possible_rooms.erase(state.right_door_type)
	
	if state.level <= 4:
		possible_rooms.erase(Rooms.RoomType.shop)
		possible_rooms.erase(Rooms.RoomType.bank)
		possible_rooms.erase(Rooms.RoomType.blood_donation)
		possible_rooms.erase(Rooms.RoomType.cursed_shop)
		possible_rooms.erase(Rooms.RoomType.chest)
		possible_rooms.erase(Rooms.RoomType.offering)
		possible_rooms.erase(Rooms.RoomType.laboratory)
	
	for room in Rooms.SINGLE_OCCURENCE_ROOMS:
		if state.previous_rooms.find(room) != -1:
			possible_rooms.erase(room)
	for room in Rooms.RoomType.values():
		if state.previous_rooms.count(room) >= 2:
			possible_rooms.erase(room)
	
	var possible_protection := Rooms.DoorProtection.values()
	possible_protection.erase(Rooms.DoorProtection.boss_key)
	if state.level <= 3:
		possible_protection = [Rooms.DoorProtection.none]
	
	if state.get_floor_number() >= 2 and state.get_floor_number() <= 4 and state.get_room_number() == 4 && Game.state.unlocked_challenge_keys.find("CHALLENGE_5") != -1:
		state.left_door_type = Rooms.RoomType.newspaper
		match state.get_floor_number():
			2:
				state.left_door_protection = Rooms.DoorProtection.key
			3:
				state.left_door_protection = Rooms.DoorProtection.rocks
			4:
				state.left_door_protection = Rooms.DoorProtection.spikes
	elif state.get_room_number() == 8 && possible_rooms.find(Rooms.RoomType.shop) != -1:
		state.left_door_type = Rooms.RoomType.shop
		state.left_door_protection = possible_protection[random_gen.randi_range(0, possible_protection.size() - 1)]
	else:
		state.left_door_type = possible_rooms[random_gen.randi_range(0, possible_rooms.size() - 1)]
		state.left_door_protection = possible_protection[random_gen.randi_range(0, possible_protection.size() - 1)]
	possible_rooms.erase(state.left_door_type)
	
	if state.get_room_number() == 8 && possible_rooms.find(Rooms.RoomType.coins) != -1:
		state.right_door_type = Rooms.RoomType.coins
	else:
		state.right_door_type = possible_rooms[random_gen.randi_range(0, possible_rooms.size() - 1)]
	
	if state.left_door_protection != Rooms.DoorProtection.none:
		state.right_door_protection = Rooms.DoorProtection.none
	else:
		state.right_door_protection = possible_protection[random_gen.randi_range(0, possible_protection.size() - 1)]
