class_name Battlefield
extends Node2D

signal on_pause()

@export var players_initial_position: Vector2
@export var rewards_position: Vector2
@export var spawn_area: Rect2
@export var arena_borders: Rect2

@onready var player_reference := preload("res://src/entities/player/Player.tscn")
@onready var heart_reference := preload("res://src/battlefield/heart/Heart.tscn")
@onready var battle_end_effect_reference := preload("res://src/effects/particles/dust_puffs/BattleEndEffect.tscn")
@onready var dash_texture := preload("res://src/battlefield/dash.png")

var players: Array[Player]
var enemies: Array[Enemy]
var decoys: Array[DecoyAssist]
var _battle_finished := false
var _reset_key_press_duration := 0.0

var _shake_amount = 0.0

func _ready():
	EventBus.create_projectile.connect(_create_projectile)
	EventBus.create_effect.connect(_create_effect)
	EventBus.add_follow_assist.connect(_add_follow_assist)
	
	EventBus.update_battlefield_ui.connect(_update_battlefield_ui)
	EventBus.shake_screen.connect(_shake_screen)
	EventBus.anim_open_door.connect(_anim_open_door)
	EventBus.finish_run.connect(_finish_run)
	EventBus.show_play_game_dialog.connect(_show_play_game_dialog)
	EventBus.enable_glitch.connect(_enable_glitch)
	EventBus.show_action_button.connect(_show_action_button)
	EventBus.hide_action_button.connect(_hide_action_button)
	
	EventBus.player_death.connect(_player_death)
	EventBus.enemy_death.connect(_enemy_death)
	
	EventBus.reroll_all_doors.connect(_reroll_all_doors)
	EventBus.on_machine_bombed.connect(_on_machine_bombed)
	
	$UI/HealthBar.visible = false
	
	var state := Game.state
	
	_init_level_number()
	_init_doors()
	
	var player_spacing := 60.0

	for player_idx in state.players.size():
		var player := player_reference.instantiate()
		player.state = state.players[player_idx]
		players.append(player)
		
		var x_delta := 0.0
		var y_delta := 0.0
		
		if state.players.size() >= 3:
			match player_idx:
				0:
					x_delta = -player_spacing / 2
					y_delta = -player_spacing
				1:
					x_delta = player_spacing / 2
					y_delta = -player_spacing
				2:
					x_delta = -player_spacing / 2
				3:
					x_delta = player_spacing / 2
		elif state.players.size() == 2:
			match player_idx:
				0:
					x_delta = -player_spacing / 2
				1:
					x_delta = player_spacing / 2
		
		player.global_position = players_initial_position + Vector2(x_delta, y_delta)
		$Entities.add_child(player)
	
	_update_battlefield_ui()
	
	if Game.state.room_type == Rooms.RoomType.boss or Game.state.room_type == Rooms.RoomType.newspaper:
		Musics.play_boss_music()
	else:
		Musics.play_dungeon_music()
	
	if Rooms.ROOMS_WITHOUT_ENEMIES.find(Game.state.room_type) != -1:
		_finish_battle(false)
	else:
		_add_enemies()
	
	_update_touch_interface()
	Input.joy_connection_changed.connect(_joy_connection_changed)
	
	match Game.state.get_floor_number():
		2:
			$LeavesParticles2D.emitting = true
		3:
			$SnowParticles2D.emitting = true
		4:
			$LightsParticles2D.emitting = true

func _joy_connection_changed(_device: int, _connected: bool):
	_update_touch_interface()

func _update_touch_interface():
	$TouchControls.visible = InputUtils.is_in_touch_mode()

func _process(_delta: float):
	if _shake_amount > 0:
		$Camera2D.offset = Vector2(randf_range(-1.0, 1.0) * _shake_amount, randf_range(-1.0, 1.0) * _shake_amount)
	
	if Input.is_action_pressed("reset"):
		_reset_key_press_duration += _delta
		if _reset_key_press_duration > 1.5:
			_reset_key_press_duration = -999.0
			Game.start_new_run(Game.state.initial_players)
			
			SceneTransition.change_scene(SceneTransition.battlefield, SceneTransition.TransitionType.slow_dissolve)

func _input(event):
	if event.is_action_released("reset"):
		_reset_key_press_duration = 0.0

func _init_level_number():
	var floor_number := Game.state.get_floor_number()
	var room_number := Game.state.get_room_number()
	$Background.texture = load("res://src/battlefield/backgrounds/floor%s_room%s.png" % [(floor_number-1) % 5 + 1, (room_number-1) % 4 + 1])
	$UI/LevelLabel.text = "%s %s-%s" % [tr("FLOOR"), floor_number, room_number]

func _init_doors():
	if Game.state.left_door_type == Rooms.RoomType.none:
		$LeftDoor.visible = false
	else:
		$LeftDoor.door_type = Game.state.left_door_type
		$LeftDoor.door_protection = Game.state.left_door_protection
	
	if Game.state.right_door_type == Rooms.RoomType.none:
		$RightDoor.visible = false
	else:
		$RightDoor.door_type = Game.state.right_door_type
		$RightDoor.door_protection = Game.state.right_door_protection

func _reroll_all_doors():
	Game.generate_doors_for_current_room(true)
	_init_doors()

func _on_machine_bombed():
	Game.state.nb_machines_bombed += 1
	if Game.state.nb_machines_bombed >= 10:
		_unlock_challenge("CHALLENGE_6")

func _anim_open_door(door: Door):
	if not _has_alive_player(): # Possible if the user died on opening the spiked door
		return
	
	$UI.visible = false
	
	var tween := get_tree().create_tween().bind_node($Camera2D)
	tween.set_parallel().set_trans(Tween.TRANS_QUAD)
	tween.tween_property($Camera2D, "zoom", Vector2(2, 2), 0.5)
	tween.tween_property($Camera2D, "global_position", door.global_position, 0.5)
	
	await tween.finished
	
	Game.set_next_level(door.door_type)
	
	# When going to a new floor, we do a longer transition
	if door.door_type == Rooms.RoomType.start:
		SceneTransition.change_scene(SceneTransition.battlefield, SceneTransition.TransitionType.slow_dissolve)
	else:
		SceneTransition.change_scene(SceneTransition.battlefield)

func _on_hit_points_changed(hit_points: int, max_hit_points: int):
	$UI/HealthBar.set_hit_points(hit_points, max_hit_points)

func _create_projectile(projectile_instance: Node):
	$Projectiles.add_child(projectile_instance)

func _create_effect(effect_instance: Node, start_pos: Vector2, above: bool, use_ysort: bool):
	effect_instance.global_position = start_pos
	if use_ysort:
		$Entities.call_deferred("add_child", effect_instance)
	elif above:
		$EffectsAbove.call_deferred("add_child", effect_instance)
	else:
		$EffectsBelow.call_deferred("add_child", effect_instance)

func _add_follow_assist(assist: BaseFollowAssist):
	# Needed to avoid a strange error when called from enemies
	$Entities.call_deferred("add_child", assist)

func get_enemy_squirrels() -> Array[ExplodingSquirrelAssist]:
	var squirrels: Array[ExplodingSquirrelAssist] = []
	for entity in $Entities.get_children():
		if entity is ExplodingSquirrelAssist and (entity as ExplodingSquirrelAssist).get_collision_layer_value(4):
			squirrels.append(entity)
	
	return squirrels

func _add_enemies():
	if Game.state.is_boss_room() or Game.state.is_newspaper_room():
		$UI/HealthBar.visible = true
	
	enemies = Enemies.get_random_enemies()
	
	for entity in $Entities.get_children():
		if entity is DecoyAssist:
			decoys.append(entity)
	
	for i in enemies.size():
		var enemy_instance = enemies[i]
		
		var x := Game.random_gen.randf_range(spawn_area.position.x, spawn_area.end.x)
		var y := Game.random_gen.randf_range(spawn_area.position.y, spawn_area.end.y)
		enemy_instance.global_position = Vector2(x, y)
		
		$Entities.add_child(enemy_instance)
		
		if Game.state.is_boss_room() or Game.state.is_newspaper_room():
			enemy_instance.hit_points_changed.connect(_on_hit_points_changed)
			
			if enemy_instance.boss_name != "":
				get_tree().paused = true
				$BossName/Label.text = enemy_instance.boss_name
				$BossName.visible = true
				$BossName/Anim.play("fade_in")
				await get_tree().create_timer(2.0).timeout
				$BossName/Anim.play("fade_out")
				get_tree().paused = false

func _has_alive_player() -> bool:
	for player in players:
		if player.is_alive():
			return true
	return false

func _player_death(_player_position: Vector2):
	if not _has_alive_player():
		_game_over(_player_position)

func _enemy_death(_enemy_position: Vector2, _dead_enemy: Enemy):
	await get_tree().create_timer(.001, false).timeout # To display the room a bit after the smoke
	
	var remaining_enemy := false
	for child in $Entities.get_children():
		if is_instance_valid(child) and child is Enemy and child != _dead_enemy:
			remaining_enemy = true
			break
	
	var gained_souls := int(Game.state.level / 2.0)
	# If we just won against a boss
	if Game.state.room_type == Rooms.RoomType.boss:
		gained_souls *= 30
	
	Game.state.total_souls += gained_souls
	Game.state.run_souls += gained_souls
	_update_battlefield_ui()
	
	if not remaining_enemy:
		_finish_battle()

func _finish_battle(show_animation = true):
	if not _battle_finished:
		_battle_finished = true
		$Glitches.visible = false
		
		if show_animation:
			EventBus.detonate_landmines.emit()
			_create_effect(battle_end_effect_reference.instantiate(), rewards_position, true, false)
		
		# Increase charges
		if Rooms.ROOMS_WITHOUT_ENEMIES.find(Game.state.room_type) == -1:
			for player in players:
				if player.state.activable_item_id > 0:
					if player.state.has_item(BigBatteryItem):
						player.state.activable_item_charges += 2
					else:
						player.state.activable_item_charges += 1
					var max_charges := ((Items.ITEMS_BY_ID[player.state.activable_item_id]).new() as BaseItem).needed_charges
					if player.state.activable_item_charges > max_charges:
						player.state.activable_item_charges = max_charges
		
		# Increase the number of killed horsemen
		if Game.state.room_type == Rooms.RoomType.newspaper or Game.state.is_first_final_boss_room():
			Game.state.nb_horsemen_killed += 1
		
		# Unlocks challenges
		var challenge_keys := [] as Array[String]
		if Game.state.room_type == Rooms.RoomType.boss and Game.state.level == 10:
			challenge_keys.append("CHALLENGE_1")
		elif Game.state.room_type == Rooms.RoomType.boss and Game.state.level == 30:
			challenge_keys.append("CHALLENGE_2")
		elif Game.state.is_first_final_boss_room() and Game.state.unlocked_challenge_keys.find("CHALLENGE_3") == -1:
			challenge_keys.append("CHALLENGE_3")
		elif Game.state.is_first_final_boss_room():
			challenge_keys.append("CHALLENGE_4")
			
			if Game.state.used_keys == 0:
				challenge_keys.append("CHALLENGE_8")
			if Game.state.used_bombs == 0:
				challenge_keys.append("CHALLENGE_9")
			
			for player in players:
				if player.state.has_item(InfiniteBombsItem):
					challenge_keys.append("CHALLENGE_10")
				if player.state.has_item(QuadShotItem):
					challenge_keys.append("CHALLENGE_11")
		elif Game.state.is_second_final_boss_room():
			challenge_keys.append("CHALLENGE_5")
			for player in players:
				if player.state.has_item(NoHeartItem):
					challenge_keys.append("CHALLENGE_12")
				if player.state.has_item(TheChallengeItem):
					challenge_keys.append("CHALLENGE_14")
			if Game.state.nb_horsemen_killed == 4:
				challenge_keys.append("CHALLENGE_13")
			if Game.state.difficulty_level == GameState.DifficultyLevel.hard:
				challenge_keys.append("CHALLENGE_15")
		
		if Game.state.total_souls >= 5000:
			challenge_keys.append("CHALLENGE_16")
		
		_unlock_challenges(challenge_keys)
		
		# Avoid letting the player quit the run and retrieve the previous state
		Game.persist_state(false)
		
		$LeftDoor.allow_interaction()
		$RightDoor.allow_interaction()
		
		var room: Node
		if Game.state.is_second_final_boss_room() and Game.state.nb_horsemen_killed == 4:
			Musics.stop_with_fade_out()
			var end_room_reference := load("res://src/battlefield/rooms/end_room/EndRoom.tscn") as PackedScene
			room = end_room_reference.instantiate()
		elif Game.state.is_first_final_boss_room() or Game.state.is_second_final_boss_room():
			var final_boss_room_reference := load("res://src/battlefield/rooms/final_boss_room/FinalBossRoom.tscn") as PackedScene
			room = final_boss_room_reference.instantiate()
		else:
			match Game.state.room_type:
				Rooms.RoomType.start:
					var start_room_reference := load("res://src/battlefield/rooms/start_room/StartRoom.tscn") as PackedScene
					room = start_room_reference.instantiate()
				Rooms.RoomType.bank:
					var bank_room_reference := load("res://src/battlefield/rooms/bank_room/BankRoom.tscn") as PackedScene
					room = bank_room_reference.instantiate()
				Rooms.RoomType.blood_donation:
					var blood_donation_room_reference := load("res://src/battlefield/rooms/blood_donation_room/BloodDonationRoom.tscn") as PackedScene
					room = blood_donation_room_reference.instantiate()
				Rooms.RoomType.shop:
					var shop_room_reference := load("res://src/battlefield/rooms/shop_room/ShopRoom.tscn") as PackedScene
					room = shop_room_reference.instantiate()
					room.players = players
				Rooms.RoomType.cursed_shop:
					var cursed_shop_room_reference := load("res://src/battlefield/rooms/cursed_shop_room/CursedShopRoom.tscn") as PackedScene
					room = cursed_shop_room_reference.instantiate()
					room.players = players
				Rooms.RoomType.chest:
					var chest_room_reference := load("res://src/battlefield/rooms/chest_room/ChestRoom.tscn") as PackedScene
					room = chest_room_reference.instantiate()
				Rooms.RoomType.newspaper:
					Musics.stop_with_fade_out()
					var newspaper_room_reference := load("res://src/battlefield/rooms/newspaper_room/NewspaperRoom.tscn") as PackedScene
					room = newspaper_room_reference.instantiate()
				Rooms.RoomType.offering:
					var offering_room_reference := load("res://src/battlefield/rooms/offering_room/OfferingRoom.tscn") as PackedScene
					room = offering_room_reference.instantiate()
				Rooms.RoomType.laboratory:
					var laboratory_room_reference := load("res://src/battlefield/rooms/laboratory_room/LaboratoryRoom.tscn") as PackedScene
					room = laboratory_room_reference.instantiate()
				_:
					var generic_item_room_reference := load("res://src/battlefield/rooms/generic_item_room/GenericItemRoom.tscn") as PackedScene
					room = generic_item_room_reference.instantiate()
					room.players = players
		
		room.global_position = rewards_position
		$Entities.add_child(room)
		
		_update_battlefield_ui()

func _show_play_game_dialog():
	var dialog_ref := load("res://src/battlefield/rooms/end_room/PlayGameDialog.tscn") as PackedScene
	add_child(dialog_ref.instantiate())

func _finish_run():
	_quit_run(false, rewards_position)

func _quit_run(is_game_over: bool, _zoom_position: Vector2):
	Game.quit_run() # We do it now so it is saved and can be ignored with an app kill
	
	Musics.play_main_menu_music()
	
	get_tree().paused = true
	
	$UI.visible = false
	
	var zoom_tween := get_tree().create_tween().bind_node($Camera2D)
	zoom_tween.set_parallel().set_trans(Tween.TRANS_QUAD)
	zoom_tween.tween_property($Camera2D, "zoom", Vector2(2, 2), 0.5)
	zoom_tween.tween_property($Camera2D, "global_position", _zoom_position, 0.5)
	await zoom_tween.finished
	
	if not is_game_over:
		$GameOver/Control/Texts/TitleLabel.text = "CONGRATS"
	$GameOver/Control/Texts/HBoxContainer/SoulsLabel.text = tr("TOTAL_SOULS_EARNED") % str(Game.state.run_souls)
	$GameOver/Control.modulate = Color.TRANSPARENT
	$GameOver.visible = true
	
	var fade_in_tween := get_tree().create_tween().bind_node($GameOver/Control)
	fade_in_tween.tween_property($GameOver/Control, "modulate", Color.WHITE, 1)
	await fade_in_tween.finished
	
	await get_tree().create_timer(2).timeout
	
	var fade_out_tween := get_tree().create_tween().bind_node($GameOver/Control/Texts)
	fade_out_tween.tween_property($GameOver/Control/Texts, "modulate", Color.TRANSPARENT, 1)
	await fade_out_tween.finished
	
	SceneTransition.change_scene(SceneTransition.main_menu, SceneTransition.TransitionType.slow_dissolve)

func _enable_glitch():
	$Glitches.visible = true

func _unlock_challenge(challenge_key: String):
	_unlock_challenges([challenge_key])

func _unlock_challenges(challenge_keys: Array[String]):
	for challenge_key in challenge_keys:
		if Game.state.unlocked_challenge_keys.find(challenge_key) == -1:
			Game.state.unlocked_challenge_keys.append(challenge_key)
			EventBus.on_item_unlocked.emit()
			
			$Unlock.modulate = Color.TRANSPARENT
			$Unlock.visible = true
			
			for unlockable in Items.LOCKED_ITEMS.keys():
				if Items.LOCKED_ITEMS[unlockable] == challenge_key:
					$Unlock/Item.texture = load("res://assets/textures/items/item_%s.png" % Items.ITEM_IDS_BY_CLASS[unlockable])
					if CharacterSelection.STARTING_ITEMS.find(unlockable) != -1:
						$Unlock/UnlockLabel.text = "CHARACTER_UNLOCKED"
					else:
						$Unlock/UnlockLabel.text = "ITEM_UNLOCKED"
					break
			
			var tweenIn := get_tree().create_tween().bind_node($Unlock)
			tweenIn.set_ease(Tween.EASE_OUT)
			tweenIn.set_trans(Tween.TRANS_QUAD)
			tweenIn.tween_property($Unlock, "modulate", Color.WHITE, 0.4)
			
			await tweenIn.finished
			
			await get_tree().create_timer(3, false).timeout
			
			var tweenOut := get_tree().create_tween().bind_node($Unlock)
			tweenOut.set_ease(Tween.EASE_OUT)
			tweenOut.set_trans(Tween.TRANS_QUAD)
			tweenOut.tween_property($Unlock, "modulate", Color.TRANSPARENT, 0.4)

func _game_over(_player_position: Vector2):
	_quit_run(true, _player_position)

func _update_battlefield_ui():
	# Update items
	$UI/CoinsLabel.text = str(Game.state.coins)
	$UI/KeysLabel.text = str(Game.state.nb_keys)
	$UI/BombsLabel.text = str(Game.state.nb_bombs)
	$UI/SoulsLabel.text = str(Game.state.total_souls)
	
	if players.size() < 4:
		$UI/Player4ActivableIndicator.visible = false
	if players.size() < 3:
		$UI/Player3ActivableIndicator.visible = false
	if players.size() < 2:
		$UI/Player2ActivableIndicator.visible = false
	
	# Update health & dashes
	for player_idx in players.size():
		var player := players[player_idx]
		
		if player.state.max_hit_points >= 16: # 8 red heart receptacles
			_unlock_challenge("CHALLENGE_7")
		
		var hearts_container: Node2D
		match player_idx:
			0: hearts_container = $UI/Player1Hearts
			1: hearts_container = $UI/Player2Hearts
			2: hearts_container = $UI/Player3Hearts
			3: hearts_container = $UI/Player4Hearts
		
		var nb_hearts := player.state.get_nb_hearts_and_shields()
		while hearts_container.get_child_count() > nb_hearts:
			hearts_container.get_child(hearts_container.get_child_count() - 1).free()
		
		var space := 62
		var pos_x_factor: int
		if player_idx == 0 || player_idx == 2:
			pos_x_factor = space
		else:
			pos_x_factor = -space
			
		var pos_y_factor: int
		if player_idx == 0 || player_idx == 1:
			pos_y_factor = space
		else:
			pos_y_factor = -space
		
		while hearts_container.get_child_count() < nb_hearts:
			var x_delta := (hearts_container.get_child_count() % 4) * pos_x_factor
			var y_delta := 0
			if hearts_container.get_child_count() >= 4:
				y_delta = pos_y_factor
			
			var new_heart := heart_reference.instantiate()
			new_heart.position = Vector2(x_delta, y_delta)
			hearts_container.add_child(new_heart)
		
		var remaining_hit_points := player.state.hit_points
		for i in nb_hearts:
			var heart: Heart = hearts_container.get_child(i)
			if remaining_hit_points >= 2:
				heart.set_full()
				remaining_hit_points -= 2
			elif remaining_hit_points >= 1:
				heart.set_half()
				remaining_hit_points -= 1
			elif i < nb_hearts - player.state.nb_fake_hearts - player.state.nb_shields:
				heart.set_empty()
			elif i < nb_hearts - player.state.nb_shields:
				heart.set_fake()
			else:
				heart.set_shield()
		
		# Now we display the correct number of dashes
		var dashes_container: Node2D
		match player_idx:
			0: dashes_container = $UI/Player1Dashes
			1: dashes_container = $UI/Player2Dashes
			2: dashes_container = $UI/Player3Dashes
			3: dashes_container = $UI/Player4Dashes
		
		for dash in dashes_container.get_children():
			dash.free()
		
		while dashes_container.get_child_count() < player.remaining_dashes:
			var x_delta := (dashes_container.get_child_count() % 4) * pos_x_factor
			
			var y_delta := 0
			if (player_idx == 0 or player_idx == 1) and nb_hearts <= 4:
				y_delta = -pos_y_factor
			
			var new_dash := Sprite2D.new()
			new_dash.texture = dash_texture
			new_dash.scale = Vector2(3, 3)
			new_dash.position = Vector2(x_delta, y_delta)
			dashes_container.add_child(new_dash)
		
		var activable_indicator: ActivableIndicator
		match player_idx:
			0: activable_indicator = $UI/Player1ActivableIndicator as ActivableIndicator
			1: activable_indicator = $UI/Player2ActivableIndicator as ActivableIndicator
			2: activable_indicator = $UI/Player3ActivableIndicator as ActivableIndicator
			3: activable_indicator = $UI/Player4ActivableIndicator as ActivableIndicator
		
		activable_indicator.set_controller_number(player.state.controller_number)
		
		if player.state.activable_item_id > 0:
			activable_indicator.set_activable_item(Items.ITEMS_BY_ID[player.state.activable_item_id], player.state.activable_item_charges)
		else:
			activable_indicator.set_activable_item(null, 0)
		
		if player.state.single_use_item_id > 0:
			activable_indicator.set_single_use_item(Items.ITEMS_BY_ID[player.state.single_use_item_id])
		else:
			activable_indicator.set_single_use_item(null)

func get_center_position() -> Vector2:
	return arena_borders.get_center()

func get_random_position(allow_player_pos = true) -> Vector2:
	var possible_pos := Vector2.ZERO
	
	while possible_pos == Vector2.ZERO:
		var margin := 30
		possible_pos = Vector2(randf_range(arena_borders.position.x + margin, arena_borders.end.x - margin), randf_range(arena_borders.position.y + margin, arena_borders.end.y - margin))
		
		if not allow_player_pos:
			for player in players:
				if player.global_position.distance_to(possible_pos) < 80:
					possible_pos = Vector2.ZERO
					break
	
	return possible_pos

func _shake_screen():
	var tweenIn := get_tree().create_tween().bind_node(self)
	tweenIn.set_ease(Tween.EASE_OUT)
	tweenIn.set_trans(Tween.TRANS_QUAD)
	tweenIn.tween_property(self, "_shake_amount", 2.0, 0.2)
	
	await tweenIn.finished
	
	var tweenOut := get_tree().create_tween().bind_node(self)
	tweenOut.set_ease(Tween.EASE_OUT)
	tweenOut.set_trans(Tween.TRANS_QUAD)
	tweenOut.tween_property(self, "_shake_amount", 0.0, 0.2)

func _on_alterning_items_timer_timeout():
	if players.any(func(p: Player): return p.state.has_item(AlterningItemsItem)):
		EventBus.altern_items.emit()

func _on_pause_button_pressed():
	on_pause.emit()

func _show_action_button():
	$TouchControls/ActionButton.visible = true

func _hide_action_button():
	$TouchControls/ActionButton.visible = false
