extends CharacterBody2D
class_name Enemy

@export var boss_name := ""
@export var speed := 50
@export var charge_speed := 350
@export var damage := 1
@export var max_hit_points := 600

var hit_points: int

const IDLE = "Idle"
const MOVE = "Move"
const CHARGE = "Charge"
const DEAD = "Dead"

var _move_state := IDLE

var _movement_direction := Vector2.ZERO
var _charge_direction := Vector2.ZERO
var _aleatory_position := Vector2.ZERO

var _long_firing_weapon: ModularWeapon
var _long_firing_direction := Vector2.ZERO
var _long_firing_reversed := false
var _long_firing_warning_ended := false

var _nb_phases := 1
var _current_phase := 1

var _patterns: Array[EnemyPattern]
var _current_pattern_idx := 0

var _poison_stack: Array[int] = []
var _fire_stack: Array[int] = []
var _bleeding_stack: Array[int] = []

@export var death_effect: PackedScene
@export var phase_end_effect: PackedScene

signal hit_points_changed(hit_points: int, max_hit_points: int)

var _poison_puddle_scene := preload("res://src/effects/fade/PoisonPuddle.tscn")
@onready var _weapon := $ModularWeapon as ModularWeapon

var _battlefield: Battlefield

func _ready():
	# More life with more player
	max_hit_points = int(max_hit_points * (_get_battlefield().players.size() / 1.5))
	
	# Adapt with difficulty level
	match Game.state.difficulty_level:
		GameState.DifficultyLevel.easy:
			max_hit_points = int(max_hit_points / 1.3)
		GameState.DifficultyLevel.hard:
			max_hit_points = int(max_hit_points * 1.3)
	
	# A bit easier in touch mode
	if InputUtils.is_in_touch_mode():
		max_hit_points = int(max_hit_points / 1.2)
	
	hit_points = max_hit_points
	
	_update_dot_ui()
	_start_current_pattern()

func _on_new_phase(_phase_number: int):
	pass

func _restart_patterns():
	_set_move_state(IDLE)
	_current_pattern_idx = 0
	_start_current_pattern()

func _start_current_pattern():
	var pattern := _patterns[_current_pattern_idx] as EnemyPattern
	pattern.on_start.call()
	$PatternTimer.start(randf_range(pattern.min_duration, pattern.max_duration))

func _on_pattern_timer_timeout():
	var finished_pattern := _patterns[_current_pattern_idx] as EnemyPattern
	finished_pattern.on_end.call()
	
	_current_pattern_idx += 1
	if _current_pattern_idx >= _patterns.size():
		_current_pattern_idx = 0
	
	_start_current_pattern()

func _process(_delta: float):
	if _move_state == DEAD:
		return
	
	if _long_firing_weapon:
		if _long_firing_direction != Vector2.ZERO:
			_long_firing_weapon.shoot(_delta, _long_firing_direction)
			_update_long_fire_direction(_delta)
		else:
			_long_firing_weapon.stop_shooting()
			_long_firing_weapon.rotation = 0.0
			_long_firing_reversed = false
		
	if _charge_direction != Vector2.ZERO:
		_movement_direction = _charge_direction
	else:
		_movement_direction = _patterns[_current_pattern_idx].get_direction.call()
		if _move_state == CHARGE:
			_charge_direction = _movement_direction
	
	if _movement_direction.x < 0:
		$Sprite2D.flip_h = true
	elif _movement_direction.x > 0:
		$Sprite2D.flip_h = false
	
	if _movement_direction != Vector2.ZERO:
		if _move_state == CHARGE:
			set_velocity(_movement_direction * charge_speed)
		else:
			set_velocity(_movement_direction * speed)
			_set_move_state(MOVE)
		move_and_slide()
	else:
		_set_move_state(IDLE)

func hurt(_damage: int, _is_critical: bool, _poison_damage: int, _fire_damage: int, _bleeding_damage: int):
	if _damage > 0:
		$BlinkAnim.play("Blink")
		
		var text_effect := TextEffect.new(str(_damage), 0.4)
		if _is_critical:
			text_effect.color = Color.ORANGE
		
		_apply_damage(_damage, text_effect)
	
	if _poison_damage > 0:
		_poison_stack.append(_poison_damage)
		if _poison_stack.size() > 10:
			_poison_stack.remove_at(0)
	
	if _fire_damage > 0:
		_fire_stack.append(_fire_damage)
		if _fire_stack.size() > 5:
			_fire_stack.remove_at(0)
	
	if _bleeding_damage > 0:
		_bleeding_stack.append(_bleeding_damage)
		if _bleeding_stack.size() > 15:
			_bleeding_stack.remove_at(0)
	
	_update_dot_ui()

func _on_dot_timer_timeout():
	if not _poison_stack.is_empty():
		var total_damage := 0
		for stack in _poison_stack:
			total_damage += stack
		
		var text_effect := TextEffect.new(str(total_damage), 0.4)
		text_effect.color = Color.DARK_GREEN
		
		_apply_damage(total_damage, text_effect)
		
		_poison_stack.remove_at(0)
	
	if not _fire_stack.is_empty():
		var total_damage := 0
		for stack in _fire_stack:
			total_damage += stack
		
		var text_effect := TextEffect.new(str(total_damage), 0.4)
		text_effect.color = Color.ORANGE_RED
		
		_apply_damage(total_damage, text_effect)
		
		_fire_stack.remove_at(0)
	
	if not _bleeding_stack.is_empty():
		var total_damage := 0
		for stack in _bleeding_stack:
			total_damage += stack
		
		var text_effect := TextEffect.new(str(total_damage), 0.4)
		text_effect.color = Color.DARK_RED
		
		_apply_damage(total_damage, text_effect)
		
		_bleeding_stack.remove_at(0)
	
	_update_dot_ui()

func _update_dot_ui():
	if $DotTimer.is_stopped() and (_poison_stack.size() > 0 or _fire_stack.size() > 0 or _bleeding_stack.size() > 0):
		$DotTimer.start()
	
	if _poison_stack.size() > 0:
		$Alterations/PoisonStack.visible = true
		$Alterations/PoisonStack/Label.text = str(_poison_stack.size())
	else:
		$Alterations/PoisonStack.visible = false
		
	if _fire_stack.size() > 0:
		$Alterations/FireStack.visible = true
		$Alterations/FireStack/Label.text = str(_fire_stack.size())
	else:
		$Alterations/FireStack.visible = false
		
	if _bleeding_stack.size() > 0:
		$Alterations/BleedingStack.visible = true
		$Alterations/BleedingStack/Label.text = str(_bleeding_stack.size())
	else:
		$Alterations/BleedingStack.visible = false
	
	var poison_y := 0
	if _fire_stack.size() == 0 and _bleeding_stack.size() == 0:
		poison_y = 16
	elif _fire_stack.size() == 0 or _bleeding_stack.size() == 0:
		poison_y = 8
	$Alterations/PoisonStack.position.y = poison_y
	
	var fire_y := 8
	if _bleeding_stack.size() == 0:
		fire_y = 16
	$Alterations/FireStack.position.y = fire_y

func _apply_damage(_damage: int, _text_effect: TextEffect):
	EventBus.emit_create_effect(_text_effect, global_position + Vector2(randi_range(-15, 15), randi_range(-15, 15)))
	
	hit_points = max(0, hit_points - _damage)
	
	if hit_points == 0:
		# there is still some phases, we refill its life
		if _nb_phases > _current_phase:
			_current_phase += 1
			_on_new_phase(_current_phase)
			
			hit_points = 1 # If we reached 0, the health bar disappears
			hit_points_changed.emit(hit_points, max_hit_points)
			hit_points = max_hit_points
			
			get_tree().paused = true
			
			var effect := phase_end_effect.instantiate()
			effect.process_mode = Node.PROCESS_MODE_ALWAYS
			EventBus.emit_create_effect(effect, global_position)
			Sounds.play_explosion($ExplosionSound, 0)
			
			await get_tree().create_timer(1).timeout
			
			var text_effect := TextEffect.new("PHASE_%s_TAUNT" % _current_phase, 1.6)
			text_effect.process_mode = Node.PROCESS_MODE_ALWAYS
			EventBus.emit_create_effect(text_effect, global_position)
			
			await get_tree().create_timer(1).timeout
			get_tree().paused = false
		# it was a boss, but it is really dead now (no more phase)
		elif _nb_phases > 1:
			$CollisionShape2D.set_deferred("disabled", true)
			hit_points_changed.emit(hit_points, max_hit_points)
			
			get_tree().paused = true
			
			var boss_death_effect_reference = load("res://src/effects/particles/dust_puffs/BattleEndEffect.tscn")
			
			var boss_death_effect = boss_death_effect_reference.instantiate()
			boss_death_effect.process_mode = Node.PROCESS_MODE_ALWAYS
			EventBus.emit_create_effect(boss_death_effect, global_position)
			
			await get_tree().create_timer(0.3).timeout
			EventBus.enemy_death.emit(global_position, self)
			visible = false
			
			var boss_death_effect2 = boss_death_effect_reference.instantiate()
			boss_death_effect2.process_mode = Node.PROCESS_MODE_ALWAYS
			EventBus.emit_create_effect(boss_death_effect2, global_position)
			Sounds.play_explosion($ExplosionSound, 0)
			
			await get_tree().create_timer(0.7).timeout
			
			get_tree().paused = false
			queue_free()
		# it was not a boss, and it is dead
		else:
			$CollisionShape2D.set_deferred("disabled", true)
			_set_move_state(DEAD)
			if death_effect != null:
				EventBus.emit_create_effect(death_effect.instantiate(), global_position)
			
			# To let the time to see the smoke above the enemy
			await get_tree().create_timer(0.3, false).timeout
			
			EventBus.enemy_death.emit(global_position, self)
			queue_free()
	
	hit_points_changed.emit(hit_points, max_hit_points)

func _set_move_state(new_move_state: String):
	if _move_state != new_move_state:
		if new_move_state == IDLE:
			$AnimTree["parameters/MovementBlend/blend_amount"] = 0
		else:
			if $Sprite2D.hframes == 4:
				$AnimTree["parameters/FrameNumbersBlend/blend_amount"] = -1
			elif $Sprite2D.hframes == 6:
				$AnimTree["parameters/FrameNumbersBlend/blend_amount"] = 0
			elif $Sprite2D.hframes == 8:
				$AnimTree["parameters/FrameNumbersBlend/blend_amount"] = 1
			$AnimTree["parameters/MovementBlend/blend_amount"] = 1
		
		_move_state = new_move_state

func _on_modular_weapon_damaged(_target: Node2D, _weapon_damage: int):
	if _target is Player:
		_target.hurt(damage) # we do not use weapon damage

func _get_nearest_player_or_decoy() -> Node2D:
	var nearest: Node2D
	
	for player in _get_battlefield().players:
		if player.is_alive() && (nearest == null || global_position.distance_to(player.global_position) < global_position.distance_to(nearest.global_position)):
			nearest = player
	
	for decoy in _get_battlefield().decoys:
		if nearest == null || global_position.distance_to(decoy.global_position) < global_position.distance_to(nearest.global_position):
			nearest = decoy
	
	return nearest

func _dont_move() -> Vector2:
	_aleatory_position = Vector2.ZERO
	return Vector2.ZERO

func _get_direction_to_nearest_player() -> Vector2:
	_aleatory_position = Vector2.ZERO
	
	var player := _get_nearest_player_or_decoy()
	if player:
		return global_position.direction_to(player.global_position)
	else:
		return Vector2.ZERO

func _get_direction_opposite_to_nearest_player() -> Vector2:
	var player := _get_nearest_player_or_decoy()
	if player:
		var distance := global_position.distance_to(player.global_position)
		# We work with a range to avoid having the enemy quickly changing direction (which looks like a bug)
		if distance < 300:
			if distance < 200 or _aleatory_position == Vector2.ZERO:
				_aleatory_position = Vector2.ZERO
				return player.global_position.direction_to(global_position)
			else:
				return _get_direction_to_aleatory_position()
		else:
			return _get_direction_to_aleatory_position()
	else:
		return Vector2.ZERO

func _get_direction_to_center() -> Vector2:
	var center_pos := _get_center_position()
	var distance := global_position.distance_to(center_pos)
	
	if _aleatory_position == Vector2.ZERO and distance < 10:
		return _get_direction_to_aleatory_position()
	if _aleatory_position != Vector2.ZERO and distance < 100:
		return _get_direction_to_aleatory_position()
	
	_aleatory_position = Vector2.ZERO
	return global_position.direction_to(center_pos)

func _get_direction_to_aleatory_position() -> Vector2:
	while _aleatory_position == Vector2.ZERO or global_position.distance_to(_aleatory_position) < 80:
		_aleatory_position = _get_random_position()
	return global_position.direction_to(_aleatory_position)

func _shoot_nearest_player(weapon := _weapon):
	var direction := _get_direction_to_nearest_player()
	if direction != Vector2.ZERO:
		weapon.shoot(0, direction)

func _bomb_nearest_player(weapon := _weapon):
	var player := _get_nearest_player_or_decoy()
	if player:
		weapon.shoot(0, player.global_position)

func _teleport_to(new_global_pos: Vector2):
	EventBus.emit_create_effect(CircleWarningEffect.new(self), new_global_pos, false)
	await get_tree().create_timer(0.8, false).timeout
	
	$AnimTree["parameters/TeleportInOneShot/request"] = AnimationNodeOneShot.ONE_SHOT_REQUEST_FIRE
	await $AnimTree.animation_finished
	global_position = new_global_pos
	$AnimTree["parameters/TeleportOutOneShot/request"] = AnimationNodeOneShot.ONE_SHOT_REQUEST_FIRE

func _get_area_borders() -> Rect2:
	return _get_battlefield().arena_borders

func _get_center_position() -> Vector2:
	return _get_battlefield().get_center_position()

func _get_left_position() -> Vector2:
	var battlefield := _get_battlefield()
	var middle_y := (battlefield.arena_borders.end.y + battlefield.arena_borders.position.y)/2
	return Vector2(battlefield.arena_borders.position.x, middle_y)

func _get_right_position() -> Vector2:
	var battlefield := _get_battlefield()
	var middle_y := (battlefield.arena_borders.end.y + battlefield.arena_borders.position.y)/2
	return Vector2(battlefield.arena_borders.end.x, middle_y)

func _get_random_position(allow_player_pos = true) -> Vector2:
	return _get_battlefield().get_random_position(allow_player_pos)

func _detonate_landmines():
	EventBus.detonate_landmines.emit()

func _teleport_aleatory():
	_teleport_to(_get_random_position())

func _teleport_to_nearest_player():
	var player := _get_nearest_player_or_decoy()
	if player:
		_teleport_to(player.global_position)

func _teleport_to_center():
	_teleport_to(_get_center_position())

func _teleport_to_left():
	_teleport_to(_get_left_position())

func _teleport_to_right():
	_teleport_to(_get_right_position())

func _move_pattern(_min_duration: float, _max_duration: float, _on_start = __, _on_end = __):
	return EnemyPattern.new(_min_duration, _max_duration, _get_direction_to_nearest_player, _on_start, _on_end)

func _wait_pattern(_duration: float, _on_start = __, _on_end = __):
	return EnemyPattern.new(_duration, _duration, func(): return Vector2.ZERO, _on_start, _on_end)

func _charge():
	if _move_state != CHARGE:
		var player := _get_nearest_player_or_decoy()
		if player:
			EventBus.emit_create_effect(LineWarningEffect.new(self, global_position.direction_to(player.global_position)), global_position, false)
			_set_move_state(CHARGE)

func _stop_charge():
	_charge_direction = Vector2.ZERO
	_set_move_state(IDLE)

func __():
	pass

func _on_poison_puddle_timer_timeout():
	var poison_puddle := _poison_puddle_scene.instantiate() as PoisonPuddle
	EventBus.emit_create_effect(poison_puddle, global_position + Vector2.DOWN * 10, false)

func add_shield_assists():
	var shield_assist_scene := load("res://src/entities/assists/shield/ShieldAssist.tscn")
	for i in 2:
		$Assists.call_deferred("add_child", shield_assist_scene.instantiate())
		await get_tree().create_timer(2.0, false).timeout

func add_flail_assist():
	var flail_assist_scene := load("res://src/entities/assists/flail/FlailAssist.tscn")
	$Assists.call_deferred("add_child", flail_assist_scene.instantiate())

func remove_assists():
	for assist in $Assists.get_children():
		var tween = get_tree().create_tween().bind_node(assist)
		tween.tween_property(assist, "scale", Vector2.ZERO, 0.1)
		tween.tween_callback(assist.queue_free)

func _add_squirrels(number := 10):
	var squirrel_scene := load("res://src/entities/assists/exploding_squirrel/ExplodingSquirrelAssist.tscn") as PackedScene
	
	for i in number:
		var squirrel := squirrel_scene.instantiate() as ExplodingSquirrelAssist
		squirrel.set_enemy(self)
		EventBus.add_follow_assist.emit(squirrel)

func _add_skullflames_on_left():
	var area_borders := _get_area_borders()
	
	var left_x := area_borders.position.x
	var top_y := area_borders.position.y
	var bottom_y := area_borders.end.y
	
	var height := bottom_y - top_y
	var distance := height / 6
	
	var skullflame_scene := load("res://src/entities/assists/skullflame/SkullflameAssist.tscn") as PackedScene
	
	for i in 7:
		var skullflame := skullflame_scene.instantiate() as BaseRandomDirectionAssist
		skullflame.global_position = Vector2(left_x, top_y + (distance * i))
		skullflame.specific_firing_direction = Vector2.RIGHT
		EventBus.add_follow_assist.emit(skullflame)

func _add_skullflames_at_random(number := 10, shoot_player := false):
	var skullflame_scene := load("res://src/entities/assists/skullflame/SkullflameAssist.tscn") as PackedScene
	
	for i in number:
		var skullflame := skullflame_scene.instantiate() as BaseRandomDirectionAssist
		var random_pos := _get_random_position()
		skullflame.global_position = random_pos
		if shoot_player:
			var player := _get_nearest_player_or_decoy()
			if player:
				skullflame.specific_firing_direction = random_pos.direction_to(player.global_position)
		EventBus.add_follow_assist.emit(skullflame)

func _fire_angle(number := 10, reverse_at_middle := false, weapon := _weapon, direction := Vector2.UP, wait_time := 0.1):
	if direction == Vector2.ZERO:
		return
	
	@warning_ignore("integer_division")
	var middle_i := number / 2
	var initial_angle := middle_i * -0.2
	if reverse_at_middle:
		initial_angle = (middle_i / 2.0) * -0.2
	
	for i in number:
		var angle := initial_angle + i * 0.2
		if reverse_at_middle and i > middle_i:
			angle = initial_angle + (middle_i * 0.2) - ((i - middle_i) * 0.2)
		weapon.shoot(0, direction.rotated(angle))
		await get_tree().create_timer(wait_time, false).timeout

func _fire_reverse_angle(number := 30, weapon := _weapon, direction := Vector2.UP, wait_time := 0.1):
	_fire_angle(number, true, weapon, direction, wait_time)

func _fire_stairs(number := 10, weapon := _weapon, direction := Vector2.UP, wait_time := 0.1):
	if direction == Vector2.ZERO:
		return
	
	@warning_ignore("integer_division")
	var middle_i := number / 2
	var initial_angle := middle_i * -0.2
	
	for i in number:
		for y in 3:
			var angle := initial_angle + i * 0.5 + y * 0.3
			weapon.shoot(0, direction.rotated(angle))
			await get_tree().create_timer(wait_time, false).timeout
		await get_tree().create_timer(wait_time * 3, false).timeout

func _long_fire(weapon := _weapon, duration := 5.0, reverse_at_middle := false, initial_direction := Vector2.UP, warning_duration := 0.9):
	_long_firing_weapon = weapon
	_long_firing_direction = initial_direction
	
	if warning_duration <= 0.0:
		_long_firing_warning_ended = true
	else:
		_long_firing_warning_ended = false
		await get_tree().create_timer(warning_duration, false).timeout
		_long_firing_warning_ended = true
	
	if reverse_at_middle:
		await get_tree().create_timer(duration/2, false).timeout
		_long_firing_reversed = true
		await get_tree().create_timer(duration/2, false).timeout
	else:
		await get_tree().create_timer(duration, false).timeout
	
	_long_firing_direction = Vector2.ZERO

func _update_long_fire_direction(_delta: float):
	if _long_firing_warning_ended:
		if _long_firing_reversed:
			_long_firing_weapon.rotate(-0.7 * _delta)
		else:
			_long_firing_weapon.rotate(0.7 * _delta)

func _get_battlefield() -> Battlefield:
	if _battlefield:
		return _battlefield
	
	_battlefield = find_parent("Battlefield")
	return _battlefield
