class_name Player
extends CharacterBody2D

const SPEED := 230
const DASH_SPEED := 800
const DASH_EFFECT_TIMER := .02

var dash_effect_timer_value := .0
var remaining_dashes := 0

@export var dash_effect: PackedScene
@export var death_effect: PackedScene

var state: PlayerState

var _enemies_in_collision: Array[Enemy] = []

var _movement_direction := Vector2.ZERO
var _shooting_direction := Vector2.ZERO
var _dash_direction := Vector2.ZERO

var _interactable_object: Interactable

const IDLE := "Idle"
const MOVE := "Move"
const DASH := "Dash"
const IDLE_DEAD := "IdleDead"
const MOVE_DEAD := "MoveDead"

var _move_state: String

var _is_invulnerable := false

var _dust_walk := preload("res://src/effects/particles/dust_puffs/DustWalkEffect.tscn")
var _poison_puddle_scene := preload("res://src/effects/fade/PoisonPuddle.tscn")

const INITIAL_NB_DASHES := 2
const INITIAL_SPEED_FACTOR := 1.0
const INITIAL_DAMAGE_FACTOR := 1.2
const INITIAL_DOT_DAMAGE_FACTOR := 1.2
const INITIAL_FIRE_RATE_FACTOR := 1.0
const INITIAL_RANGE_FACTOR := 1.0
const INITIAL_CRITICAL_CHANCE := 0.1
const INITIAL_CRITICAL_DAMAGE_FACTOR := 1.5

var _items: Array[BaseItem] = []
var _activable_item: BaseItem
var _single_use_item: BaseItem

@onready var _weapon := $Weapon/ModularWeapon as ModularWeapon
@onready var _ghost_weapon := $Weapon/GhostWeapon as ModularWeapon
@onready var _bomb_weapon := $BombWeapon as ModularWeapon

var _battlefield: Battlefield

func get_nb_dashes() -> int:
	var value := INITIAL_NB_DASHES
	for item in _items:
		value = item.get_nb_dashes(value, self)
	
	return max(0, min(4, value))

func get_speed_factor() -> float:
	if not is_alive():
		return INITIAL_SPEED_FACTOR
	
	var value := INITIAL_SPEED_FACTOR
	var has_stat_down_defense := false
	for item in _items:
		if item is StatDownDefenseItem:
			has_stat_down_defense = true
		var new_value = item.get_speed_factor(value, self)
		if not has_stat_down_defense or new_value > value:
			value = new_value
	
	return max(0.1, value)

func get_damage_factor() -> float:
	if not is_alive():
		return INITIAL_DAMAGE_FACTOR
	
	var value := INITIAL_DAMAGE_FACTOR
	var has_stat_down_defense := false
	for item in _items:
		if item is StatDownDefenseItem:
			has_stat_down_defense = true
		var new_value = item.get_damage_factor(value, self)
		if not has_stat_down_defense or new_value > value:
			value = new_value
	
	return max(0.1, value)

func get_dot_damage_factor() -> float:
	if not is_alive():
		return INITIAL_DOT_DAMAGE_FACTOR
	
	var value := INITIAL_DOT_DAMAGE_FACTOR
	var has_stat_down_defense := false
	for item in _items:
		if item is StatDownDefenseItem:
			has_stat_down_defense = true
		var new_value = item.get_dot_damage_factor(value, self)
		if not has_stat_down_defense or new_value > value:
			value = new_value
	
	return max(0.1, value)

func get_fire_rate_factor() -> float:
	if not is_alive():
		return INITIAL_FIRE_RATE_FACTOR
	
	var value := INITIAL_FIRE_RATE_FACTOR
	var has_stat_down_defense := false
	for item in _items:
		if item is StatDownDefenseItem:
			has_stat_down_defense = true
		var new_value = item.get_fire_rate_factor(value, self)
		if not has_stat_down_defense or new_value > value:
			value = new_value
	
	return max(0.1, min(4.0, value))

func get_range_factor() -> float:
	if not is_alive():
		return INITIAL_RANGE_FACTOR
	
	var value := INITIAL_RANGE_FACTOR
	var has_stat_down_defense := false
	for item in _items:
		if item is StatDownDefenseItem:
			has_stat_down_defense = true
		var new_value = item.get_range_factor(value, self)
		if not has_stat_down_defense or new_value > value:
			value = new_value
	
	return max(0.1, value)

func get_critical_chance() -> float:
	if not is_alive():
		return INITIAL_CRITICAL_CHANCE
	
	var value := INITIAL_CRITICAL_CHANCE
	var has_stat_down_defense := false
	for item in _items:
		if item is StatDownDefenseItem:
			has_stat_down_defense = true
		var new_value = item.get_critical_chance(value, self)
		if not has_stat_down_defense or new_value > value:
			value = new_value
	
	return max(0.1, value)

func get_critical_damage_factor() -> float:
	if not is_alive():
		return INITIAL_CRITICAL_DAMAGE_FACTOR
	
	var value := INITIAL_CRITICAL_DAMAGE_FACTOR
	var has_stat_down_defense := false
	for item in _items:
		if item is StatDownDefenseItem:
			has_stat_down_defense = true
		var new_value = item.get_critical_damage_factor(value, self)
		if not has_stat_down_defense or new_value > value:
			value = new_value
	
	return max(0.1, value)

func does_apply_poison() -> bool:
	if not is_alive():
		return false
	
	var value := false
	for item in _items:
		value = item.does_apply_poison(value, self)
	
	return value

func does_apply_fire() -> bool:
	if not is_alive():
		return false
	
	var value := false
	for item in _items:
		value = item.does_apply_fire(value, self)
	
	return value

func does_apply_bleeding() -> bool:
	if not is_alive():
		return false
	
	var value := false
	for item in _items:
		value = item.does_apply_bleeding(value, self)
	
	return value

func find_item(item_class: GDScript) -> BaseItem:
	for item in _items:
		if item.get_script() == item_class:
			return item
	return null

func is_alive():
	return state.is_alive()

func _get_commands():
	return {
		"move_up": "p%d_move_up" % state.controller_number,
		"move_down": "p%d_move_down" % state.controller_number,
		"move_left": "p%d_move_left" % state.controller_number,
		"move_right": "p%d_move_right" % state.controller_number,
		"shoot_up": "p%d_shoot_up" % state.controller_number,
		"shoot_down": "p%d_shoot_down" % state.controller_number,
		"shoot_left": "p%d_shoot_left" % state.controller_number,
		"shoot_right": "p%d_shoot_right" % state.controller_number,
		"dash": "p%d_dash" % state.controller_number,
		"interaction": "p%d_interaction" % state.controller_number,
		"bomb": "p%d_bomb" % state.controller_number,
		"activable": "p%d_activable" % state.controller_number,
		"single_use": "p%d_single_use" % state.controller_number,
	}

func _ready():
	_weapon.player = self
	# It's not clear why we need to add the straight motion in code, when done in the editor motions stay between runs
	_weapon.add_motion("res://src/weapon/ProjectileMotions/Motions/StraightMotion.tres")
	
	_bomb_weapon.player = self
	_ghost_weapon.player = self
	remaining_dashes = get_nb_dashes()
	for item_id in state.item_ids:
		var item = Items.ITEMS_BY_ID[item_id].new() as BaseItem
		item.modify_weapon(_weapon)
		item.modify_bomb_weapon(_bomb_weapon)
		item.add_assist(self)
		if _move_state == IDLE:
			item.on_player_idle_begin(self)
		if item.keep_in_item_list:
			_items.append(item)
		
		if state.activable_item_id == item_id:
			_activable_item = item
		if state.single_use_item_id == item_id:
			_single_use_item = item
	
	# If the activable/single_use item wasn't in the item list (ie. keep_in_item_list = false), then we instanciate a new one
	if state.activable_item_id != 0 and _activable_item == null:
		_activable_item = Items.ITEMS_BY_ID[state.activable_item_id].new()
	if state.single_use_item_id != 0 and _single_use_item == null:
		_single_use_item = Items.ITEMS_BY_ID[state.single_use_item_id].new()
	
	if state.color_number != 1:
		$Character.texture = load("res://src/entities/player/character_variant_%s.png" % state.color_number)
		$Ghost.texture = load("res://src/entities/player/ghost_variant_%s.png" % state.color_number)

	if is_alive():
		_set_move_state(IDLE)
	else:
		_set_move_state(IDLE_DEAD)

func _auto_aim_shooting_direction() -> Vector2:
	if Game.state.firing_assistance == GameState.FiringAssistance.none:
		return _shooting_direction
	
	var shoot_angle := _shooting_direction.angle()
	
	# If we don't have QuadShotItem, we help the user by correcting a bit the angle
	if not state.has_item(QuadShotItem):
		var sorted_enemies: Array[Node2D] = []
		sorted_enemies.append_array(_get_battlefield().enemies.filter(func(e): return is_instance_valid(e)))
		sorted_enemies.append_array(_get_battlefield().get_enemy_squirrels())
		sorted_enemies.sort_custom(func(a, b): return global_position.distance_to(a.global_position) < global_position.distance_to(b.global_position))
		
		for enemy in sorted_enemies:
			var angle_to_enemy := global_position.angle_to_point(enemy.global_position)
			if abs(MathUtils.angle_difference(shoot_angle, angle_to_enemy)) < 0.26: # 0.26 rad = around 15 degree
				return global_position.direction_to(enemy.global_position);
	
	return _shooting_direction

func get_nearest_enemy() -> Node2D:
	var nearest_enemy: Node2D
	
	for enemy in _get_battlefield().enemies:
		if is_instance_valid(enemy) and enemy.hit_points > 0 and (nearest_enemy == null or global_position.distance_to(enemy.global_position) < global_position.distance_to(nearest_enemy.global_position)):
			nearest_enemy = enemy
	
	for squirrel in _get_battlefield().get_enemy_squirrels():
		if nearest_enemy == null or global_position.distance_to(squirrel.global_position) < global_position.distance_to(nearest_enemy.global_position):
			nearest_enemy = squirrel
	
	return nearest_enemy

func _process(delta: float):
	_movement_direction = Vector2.ZERO
	_shooting_direction = Vector2.ZERO
	
	var commands = _get_commands()
	
	# Update alive/dead state if needed
	if is_alive() and (_move_state == IDLE_DEAD or _move_state == MOVE_DEAD):
		_set_move_state(IDLE)
	elif not is_alive() and _move_state != IDLE_DEAD and _move_state != MOVE_DEAD:
		$Sounds/DeathSound.play()
		_set_move_state(IDLE_DEAD)
		EventBus.player_death.emit(global_position)
		EventBus.emit_create_effect(death_effect.instantiate(), global_position)
	
	# Update the number of remaining dashes if needed
	var nb_dashes := get_nb_dashes()
	if remaining_dashes < nb_dashes && $Timers/DashRestoreTimer.is_stopped():
		remaining_dashes = nb_dashes
		EventBus.update_battlefield_ui.emit()
		
	_update_alteration_icons()
	
	if _move_state == DASH:
		# create dash effect
		if dash_effect_timer_value >= DASH_EFFECT_TIMER:
			dash_effect_timer_value = 0
			if dash_effect != null:
				var dash_effect_instance = dash_effect.instantiate()
				dash_effect_instance.flip_h = $Character.flip_h
				dash_effect_instance.texture = $Character.texture
				EventBus.emit_create_effect(dash_effect_instance, global_position, false, true)
		else:
			dash_effect_timer_value += delta
		
		set_velocity(_dash_direction.normalized() * DASH_SPEED)
		move_and_slide()
	else:
		_movement_direction = Input.get_vector(commands.move_left, commands.move_right, commands.move_up, commands.move_down).normalized()
		if _movement_direction != Vector2.ZERO:
			if _move_state == IDLE:
				_set_move_state(MOVE)
			elif _move_state == IDLE_DEAD:
				_set_move_state(MOVE_DEAD)
		elif _move_state == MOVE:
			_set_move_state(IDLE)
		elif _move_state == MOVE_DEAD:
			_set_move_state(IDLE_DEAD)
		
		_shooting_direction = Input.get_vector(commands.shoot_left, commands.shoot_right, commands.shoot_up, commands.shoot_down).normalized()
		
		if InputUtils.is_in_touch_mode() or (_shooting_direction == Vector2.ZERO and Game.state.firing_assistance == GameState.FiringAssistance.auto):
			var nearest_enemy := get_nearest_enemy()
			if nearest_enemy:
				_shooting_direction = global_position.direction_to(nearest_enemy.global_position)
		
		set_velocity(_movement_direction * SPEED * get_speed_factor())
		move_and_slide()
		
		_set_player_sprite()
		_set_weapon_sprite()
		
		if _shooting_direction != Vector2.ZERO and not _items.any(func(item: BaseItem): return item.does_prevent_firing()):
			# We add the auto aim and remove the shooting_direction angle because the weapon is already rotated in that direction
			var real_weapon_direction := _auto_aim_shooting_direction().rotated(-_shooting_direction.angle())
			if _move_state != IDLE_DEAD and _move_state != MOVE_DEAD:
				_weapon.shoot(delta, real_weapon_direction)
			else:
				_ghost_weapon.shoot(delta, real_weapon_direction)
		else:
			_weapon.stop_shooting()
			_ghost_weapon.stop_shooting()

func _unhandled_input(_event: InputEvent):
	var commands = _get_commands()
	
	if _move_state == MOVE and _event.is_action_pressed(commands.dash) and _movement_direction != Vector2.ZERO and remaining_dashes > 0:
		_dash()
	elif _event.is_action_pressed(commands.interaction) and _interactable_object != null:
		_interactable_object.interact(self)
	elif _event.is_action_pressed(commands.bomb):
		var has_infinite_bombs := state.has_item(InfiniteBombsItem)
		if has_infinite_bombs or Game.state.nb_bombs > 0:
			Sounds.play_drop()
			if not has_infinite_bombs:
				Game.state.nb_bombs -= 1
				EventBus.update_battlefield_ui.emit()
			_bomb_weapon.shoot(0, Vector2.ZERO)
			Game.state.used_bombs += 1
		else:
			EventBus.emit_create_effect(TextEffect.new("NO_BOMB"), global_position)
	elif _event.is_action_pressed(commands.activable):
		if _activable_item:
			if state.activable_item_charges >= _activable_item.needed_charges:
				Sounds.play_interaction()
				state.activable_item_charges = 0
				_activable_item.on_activation(self, _weapon)
				EventBus.update_battlefield_ui.emit()
	elif _event.is_action_pressed(commands.single_use):
		if _single_use_item:
			Sounds.play_interaction()
			_single_use_item.on_activation(self, _weapon)
			_single_use_item = null
			state.single_use_item_id = 0
			EventBus.update_battlefield_ui.emit()

func _set_move_state(new_move_state: String):
	if new_move_state != _move_state:
		$Timers/DustTimer.stop()
		
		match(new_move_state):
			IDLE:
				$Character.visible = true
				$Ghost.visible = false
				call_deferred("set_collision_layer_value", 2, true)
				$HitBox/CollisionShape2D.set_deferred("disabled", false)
				$AnimTree["parameters/MovementBlend/blend_amount"] = 0
				$AnimTree["parameters/DashBlend/blend_amount"] = 0
				$AnimTree["parameters/DeathBlend/blend_amount"] = 0
				
				for item in _items:
					item.on_player_idle_begin(self)
			MOVE:
				$Timers/DustTimer.start()
				$AnimTree["parameters/MovementBlend/blend_amount"] = 1
				$AnimTree["parameters/DashBlend/blend_amount"] = 0
				$AnimTree["parameters/DeathBlend/blend_amount"] = 0
			DASH:
				call_deferred("set_collision_layer_value", 2, false) # We disable collision with bullets/toxic trails
				$HitBox/CollisionShape2D.set_deferred("disabled", true) # We disable collision with enemies
				$AnimTree["parameters/MovementBlend/blend_amount"] = 0
				$AnimTree["parameters/DashBlend/blend_amount"] = 1
				$AnimTree["parameters/DeathBlend/blend_amount"] = 0
				
				for item in _items:
					item.on_player_dash_begin(self)
			IDLE_DEAD:
				$Character.visible = false
				$Ghost.visible = true
				call_deferred("set_collision_layer_value", 2, false)
				$HitBox/CollisionShape2D.set_deferred("disabled", false) # We need to keep the hitbox for being able to interact with objects while dead
				$AnimTree["parameters/DeadMovementBlend/blend_amount"] = 0
				$AnimTree["parameters/DeathBlend/blend_amount"] = 1
			MOVE_DEAD:
				$Timers/DustTimer.start()
				$AnimTree["parameters/DeadMovementBlend/blend_amount"] = 1
				$AnimTree["parameters/DeathBlend/blend_amount"] = 1
		
		if _move_state == IDLE:
			for item in _items:
				item.on_player_idle_end(self)
		elif _move_state == DASH:
			for item in _items:
				item.on_player_dash_end(self)
		
		_move_state = new_move_state

func _set_player_sprite():
	var direction: Vector2
	if _shooting_direction != Vector2.ZERO:
		direction = _shooting_direction
	else:
		direction = _movement_direction
		
	if direction.x < 0:
		$Character.flip_h = true
		$Ghost.flip_h = true
	elif direction.x > 0:
		$Character.flip_h = false
		$Ghost.flip_h = false

func _update_alteration_icons():
	for item in _items:
		var icon := _find_alteration_icon(item)
		if is_alive() and item.is_active(self):
			if not icon:
				$AlterationsContainer.add_child(AlterationIcon.new(item))
		elif icon and not icon.will_auto_destroyed:
			$AlterationsContainer.remove_child(icon)

func _find_alteration_icon(item: BaseItem) -> AlterationIcon:
	for icon in $AlterationsContainer.get_children():
		if icon.item.get_script() == item.get_script():
			return icon
	return null

func _set_weapon_sprite():
	var direction: Vector2
	if _shooting_direction != Vector2.ZERO:
		direction = _shooting_direction
	else:
		direction = _movement_direction
	
	if direction != Vector2.ZERO:
		$Weapon.rotate(direction.angle() - $Weapon.rotation)

func _dash():
	_set_move_state(DASH)
	_dash_direction = _movement_direction
	$Sounds/DashSound.pitch_scale = 1 + randf() / 3.3
	$Sounds/DashSound.play()
	$Timers/DashTimer.start()
	_weapon.stop_shooting()
	
	remaining_dashes -= 1
	EventBus.update_battlefield_ui.emit()
	
	if $Timers/DashRestoreTimer.is_stopped():
		$Timers/DashRestoreTimer.start()

func _on_dash_timer_timeout():
	_stop_dash()

func _stop_dash():
	_set_move_state(IDLE)
	_on_invulnerability_timer_timeout()

func _on_dash_restore_timer_timeout():
	remaining_dashes += 1
	EventBus.update_battlefield_ui.emit()
	
	if remaining_dashes < get_nb_dashes():
		$Timers/DashRestoreTimer.start()

func _on_dust_timer_timeout():
	var dust := _dust_walk.instantiate()
	dust.position = Vector2(0, 5)
	add_child(dust)

func add_coins(nb_coins: int):
	if state.has_item(MoreCoinsItem):
		@warning_ignore("narrowing_conversion")
		nb_coins = nb_coins * 1.3
	
	Game.state.coins += nb_coins
	EventBus.emit_create_effect(TextEffect.new("+%s %s" % [nb_coins, tr("COINS")]), global_position)

func add_item(item: GDScript):
	var new_item := item.new() as BaseItem
	
	new_item.on_obtain(state)
	new_item.on_obtain_in_run(self)
	new_item.modify_weapon(_weapon)
	new_item.modify_bomb_weapon(_bomb_weapon)
	new_item.add_assist(self)
	if _move_state == IDLE:
		new_item.on_player_idle_begin(self)
	
	if new_item.single_use:
		_single_use_item = new_item
		state.single_use_item_id = Items.ITEM_IDS_BY_CLASS[item]
	elif new_item.is_activable:
		_activable_item = new_item
		state.activable_item_id = Items.ITEM_IDS_BY_CLASS[item]
		state.activable_item_charges = new_item.needed_charges
	
	if new_item.keep_in_item_list:
		_items.append(new_item)
		state.item_ids.append(Items.ITEM_IDS_BY_CLASS[item])
	
	EventBus.update_battlefield_ui.emit()

func remove_item(item_to_remove: GDScript):
	state.item_ids.erase(Items.ITEM_IDS_BY_CLASS[item_to_remove])
	for item in _items:
		if item.get_script() == item_to_remove:
			_items.erase(item)

func add_assist(assist_scene: PackedScene):
	var assist := assist_scene.instantiate()
	if assist is BaseOrbitAssist:
		assist.set_player(self)
		$Assists.add_child(assist)
	elif assist is BaseFollowAssist:
		assist.set_player(self)
		EventBus.add_follow_assist.emit(assist)

func hurt(damage: int, bypass_shields := false, bypass_invulnerability := false, bypass_evasion := false, display_hurt_message := false, ignore_summoner := false):
	if is_alive() && (!_is_invulnerable or bypass_invulnerability):
		$AnimTree["parameters/BlinkOneShot/request"] = AnimationNodeOneShot.ONE_SHOT_REQUEST_FIRE
		start_invulnerability()
		
		var evade_item := find_item(EvadeFirstBlowItem) as EvadeFirstBlowItem
		if not bypass_evasion and evade_item and evade_item.should_evade_first_blow:
			evade_item.should_evade_first_blow = false
			EventBus.emit_create_effect(TextEffect.new("EVADED"), global_position)
		else:
			EventBus.shake_screen.emit()
			var joypad_id := InputUtils.get_joypad_id(state.controller_number)
			if joypad_id != -1:
				Input.start_joy_vibration(joypad_id, 0.5, 0.5, 0.2)
			if display_hurt_message:
				EventBus.emit_create_effect(TextEffect.new("DAMAGED"), global_position)
			
			if state.nb_shields > 0 and not bypass_shields:
				state.nb_shields -= 1
			else:
				if state.nb_fake_hearts > 0:
					state.nb_fake_hearts -= 1
				else:
					state.hit_points = max(0, state.hit_points - damage)
				
				# small chance to add an assist when Summoner
				for i in damage:
					state.total_received_damages += 1
					if not ignore_summoner and state.total_received_damages % SummonerItem.NB_DAMAGE_TO_SUMMON == 0 and state.has_item(SummonerItem):
						call_deferred("add_item", SkullflameAssistItem)
			
			if is_alive():
				$Sounds/HurtSound.play()
			else:
				if state.has_item(ReviveItem) and not state.revive_used:
					state.revive_used = true
					$Sounds/ReviveSound.play()
					
					if state.has_item(NoHeartItem):
						state.add_shield()
						state.add_shield()
					else:
						state.fill_red_heart() # This one will add an heart receptacle if needed
						state.fill_red_heart()
					EventBus.emit_create_effect(TextEffect.new("REVIVED"), global_position)
				else:
					state.remove_heart_receptacle() # We loose 1 heart receptacle each time we die
			
			EventBus.update_battlefield_ui.emit()
		
		# We persist the state so that it is not possible to quit the run and retrieve the lost hearts/shields
		Game.persist_state(false)

func start_invulnerability(duration: float = 0.8):
	$Timers/InvulnerabilityTimer.start(duration)
	_is_invulnerable = true

func add_temp_alteration_icon(item: BaseItem, duration: float):
	var alteration_icon := AlterationIcon.new(item)
	alteration_icon.will_auto_destroyed = true
	$AlterationsContainer.add_child(alteration_icon)
	
	await get_tree().create_timer(duration - 1.6, false).timeout

	var transparent_color := Color.WHITE
	transparent_color.a = 0.2

	alteration_icon.modulate = transparent_color
	await get_tree().create_timer(0.2, false).timeout
	alteration_icon.modulate = Color.WHITE
	await get_tree().create_timer(0.2, false).timeout
	alteration_icon.modulate = transparent_color
	await get_tree().create_timer(0.2, false).timeout
	alteration_icon.modulate = Color.WHITE
	await get_tree().create_timer(0.2, false).timeout
	alteration_icon.modulate = transparent_color
	await get_tree().create_timer(0.2, false).timeout
	alteration_icon.modulate = Color.WHITE
	await get_tree().create_timer(0.2, false).timeout
	alteration_icon.modulate = transparent_color
	await get_tree().create_timer(0.2, false).timeout
	alteration_icon.modulate = Color.WHITE
	await get_tree().create_timer(0.2, false).timeout

	$AlterationsContainer.remove_child(alteration_icon)

func _on_invulnerability_timer_timeout():
	_is_invulnerable = false
	
	if not _enemies_in_collision.is_empty():
		hurt(_enemies_in_collision.front().damage)

func _on_hit_box_area_entered(area: Area2D):
	if area is Interactable and not area.is_consumed:
		if area.is_interactable():
			EventBus.show_action_button.emit()
		_interactable_object = area
		_interactable_object.show_description(state.controller_number)

func _on_hit_box_area_exited(area: Area2D):
	if area is Interactable:
		EventBus.hide_action_button.emit()
		area.hide_description()
		_interactable_object = null

func _on_hit_box_body_entered(body: Node2D):
	if body is Enemy:
		_enemies_in_collision.append(body)
		hurt(body.damage)

func _on_hit_box_body_exited(body: Node2D):
	if body is Enemy:
		_enemies_in_collision.erase(body)

func _on_modular_weapon_damaged(_target: Node2D, _damage: int):
	if _target is Destroyable:
		_target.destroy(self)
	elif _target is ExplodingSquirrelAssist:
		_target.explode()
	elif _target is Enemy:
		var damage_factor := INITIAL_DAMAGE_FACTOR
		var critical_chance := INITIAL_CRITICAL_CHANCE
		var critical_damage_factor := INITIAL_CRITICAL_DAMAGE_FACTOR
		
		if is_alive():
			damage_factor = get_damage_factor()
			critical_chance = get_critical_chance()
			critical_damage_factor = get_critical_damage_factor()
			
		var real_damage := _damage * damage_factor * randf_range(0.9, 1.1)
		var is_critical := false
		
		if randf_range(0., 1.) <= critical_chance:
			real_damage *= critical_damage_factor
			is_critical = true
		
		var poison_damage := 0
		var fire_damage := 0
		var bleeding_damage := 0
		if is_alive():
			var dot_damage_factor := get_dot_damage_factor()
			if does_apply_poison():
				poison_damage = int(20.0 * dot_damage_factor)
			if does_apply_fire():
				fire_damage = int(30.0 * dot_damage_factor)
			if does_apply_bleeding():
				bleeding_damage = int(15.0 * dot_damage_factor)
			
			if is_critical and state.has_item(BleedOnCritItem):
				for i in 5:
					_target.hurt(0, 0, 0, 0, int(15.0 * dot_damage_factor))
		
		_target.hurt(real_damage, is_critical, poison_damage, fire_damage, bleeding_damage)

func start_poison_puddle_timer():
	$Timers/PoisonPuddleTimer.start()

func _on_poison_puddle_timer_timeout():
	var poison_puddle := _poison_puddle_scene.instantiate() as PoisonPuddle
	poison_puddle.emitting_player = self
	EventBus.emit_create_effect(poison_puddle, global_position + Vector2.DOWN * 10, false)

func _get_battlefield() -> Battlefield:
	if _battlefield:
		return _battlefield
	
	_battlefield = find_parent("Battlefield")
	return _battlefield
