extends BaseMenuScreen

enum SelectedMenu {display, language, music, reset}

var _selected_menu := SelectedMenu.display

func _ready():
	_update_display_text()
	_update_language_text()
	_update_music_text()
	
	_update_touch_interface()
	Input.joy_connection_changed.connect(_joy_connection_changed)

func _joy_connection_changed(_device: int, _connected: bool):
	_update_touch_interface()

func _update_touch_interface():
	var is_in_touch_mode := InputUtils.is_in_touch_mode()
	
	$DisplayCatLabel.visible = not is_in_touch_mode
	$DisplayLabel.visible = not is_in_touch_mode
	$MenuSelector.visible = not is_in_touch_mode
	$LanguageTouchArrowSelector.visible = is_in_touch_mode
	$MusicTouchArrowSelector.visible = is_in_touch_mode

func _unhandled_input(_event: InputEvent):
	super(_event) # needed to detect menu actions
	
	if $ResetConfirmationDialog.visible:
		if _is_cancel_pressed:
			_on_back_button_on_pressed()
		elif _is_interaction_pressed:
			_on_ok_button_on_pressed()
	else:
		if _is_menu_up_pressed:
			match _selected_menu:
				SelectedMenu.display:
					_set_selected_menu(SelectedMenu.reset)
				SelectedMenu.language:
					if $DisplayCatLabel.visible:
						_set_selected_menu(SelectedMenu.display)
					else:
						_set_selected_menu(SelectedMenu.reset)
				SelectedMenu.music:
					_set_selected_menu(SelectedMenu.language)
				SelectedMenu.reset:
					_set_selected_menu(SelectedMenu.music)
		if _is_menu_down_pressed:
			match _selected_menu:
				SelectedMenu.display:
					_set_selected_menu(SelectedMenu.language)
				SelectedMenu.language:
					_set_selected_menu(SelectedMenu.music)
				SelectedMenu.music:
					_set_selected_menu(SelectedMenu.reset)
				SelectedMenu.reset:
					if $DisplayCatLabel.visible:
						_set_selected_menu(SelectedMenu.display)
					else:
						_set_selected_menu(SelectedMenu.language)
		if _is_interaction_pressed:
			match _selected_menu:
				SelectedMenu.display:
					_switch_display()
				SelectedMenu.language:
					_switch_language()
				SelectedMenu.music:
					_switch_music()
				SelectedMenu.reset:
					_ask_reset_progression()
		if _is_menu_right_pressed or _is_menu_left_pressed:
			match _selected_menu:
				SelectedMenu.display:
					_switch_display()
				SelectedMenu.language:
					_switch_language()
				SelectedMenu.music:
					_switch_music()
		
		if _is_cancel_pressed:
			_back()

func _back():
	Sounds.play_interaction()
	SceneTransition.change_scene(SceneTransition.main_menu)

func _switch_display():
	Sounds.play_interaction()
	if DisplayServer.window_get_mode() == DisplayServer.WINDOW_MODE_EXCLUSIVE_FULLSCREEN: 
		DisplayServer.window_set_mode(DisplayServer.WINDOW_MODE_WINDOWED)
	else:
		DisplayServer.window_set_mode(DisplayServer.WINDOW_MODE_EXCLUSIVE_FULLSCREEN)
	
	if DisplayServer.window_get_mode() == DisplayServer.WINDOW_MODE_EXCLUSIVE_FULLSCREEN: 
		Input.mouse_mode = Input.MOUSE_MODE_HIDDEN
	else:
		Input.mouse_mode = Input.MOUSE_MODE_VISIBLE
	
	_update_display_text()

func _switch_language():
	Sounds.play_interaction()
	var current_locale := TranslationServer.get_locale()
	if current_locale == "fr":
		TranslationServer.set_locale("en")
	else:
		TranslationServer.set_locale("fr")
	
	Game.state.language_code = TranslationServer.get_locale()
	Game.persist_state()
	_update_language_text()

func _switch_music():
	Sounds.play_interaction()
	Game.state.musics_enabled = !Game.state.musics_enabled
	Game.persist_state()
	_update_music_text()
	
	if Game.state.musics_enabled:
		Musics.play_main_menu_music()
	else:
		Musics.stop_immediately()

func _ask_reset_progression():
	Sounds.play_interaction()
	$ResetConfirmationDialog.visible = true

func _reset_progression():
	Game.state = GameState.new()
	Game.persist_state()

func _set_selected_menu(_new_selected_menu: SelectedMenu):
	_selected_menu = _new_selected_menu
	
	var tween := get_tree().create_tween().bind_node($MenuSelector)
	tween.set_parallel().set_trans(Tween.TRANS_QUAD)
	
	match _selected_menu:
		SelectedMenu.display:
			tween.tween_property($MenuSelector, "global_position", $DisplayLabel.get_global_rect().get_center(), 0.1)
		SelectedMenu.language:
			tween.tween_property($MenuSelector, "global_position", $LanguageLabel.get_global_rect().get_center(), 0.1)
		SelectedMenu.music:
			tween.tween_property($MenuSelector, "global_position", $MusicLabel.get_global_rect().get_center(), 0.1)
		SelectedMenu.reset:
			tween.tween_property($MenuSelector, "global_position", $ResetLabel.get_global_rect().get_center(), 0.1)

func _update_display_text():
	if DisplayServer.window_get_mode() == DisplayServer.WINDOW_MODE_EXCLUSIVE_FULLSCREEN: 
		$DisplayLabel.text = "FULLSCREEN"
	else:
		$DisplayLabel.text = "WINDOWED"

func _update_language_text():
	var current_locale := TranslationServer.get_locale()
	if current_locale == "en":
		$LanguageLabel.text = "ENGLISH"
	else:
		$LanguageLabel.text = "FRENCH"

func _update_music_text():
	if Game.state.musics_enabled:
		$MusicLabel.text = "MUSIC_ENABLED"
	else:
		$MusicLabel.text = "MUSIC_DISABLED"

func _on_back_button_on_pressed():
	Sounds.play_interaction()
	$ResetConfirmationDialog.visible = false

func _on_ok_button_on_pressed():
	Sounds.play_interaction()
	$ResetConfirmationDialog.visible = false
	_reset_progression()
