# Main class to manage, configure and control projectile emitters from.
class_name ModularWeapon
extends Node2D


signal damaged(target: Node2D, amount: int)

@export var emitter_configuration: PackedScene : set = _set_emitter_configuration
@export var projectile_emitter: PackedScene : set = _set_emitter
@export var projectile_motions: Array[Resource] = []
@export var projectile_impacts: Array[Resource] = []
@export var max_distance := 300.0
@export var warning_radius := 60.0
@export var projectiles_per_second := 3.0
@export var can_destroy_objects := false

var player: Player
var additionnal_volume_db_adjustment := 0.0

func shoot(delta: float, shooting_direction: Vector2):
	for emitter in get_children():
		emitter.shoot(delta, shooting_direction)

func stop_shooting():
	for emitter in get_children():
		emitter.stop_shooting()

# Appends the specified motion to the projectile motions array. Checks for and prevents
# duplicates by default, unless the second parameter is true.
func add_motion(motion_ressource_path: String, allows_duplicates := false):
	var new_motion := ResourceLoader.load(motion_ressource_path, "", ResourceLoader.CACHE_MODE_IGNORE)
	if not allows_duplicates:
		var has_motion := false
		for motion in projectile_motions:
			has_motion = new_motion.get_script() == motion.get_script()
			if has_motion:
				return
	projectile_motions.append(new_motion)

# Appends the specified impact to the projectile impacts array. Checks for and prevents
# duplicates by default, unless the second parameter is true.
func add_impact(impact_ressource_path: String, allows_duplicates := false):
	var new_impact := ResourceLoader.load(impact_ressource_path, "", ResourceLoader.CACHE_MODE_IGNORE)
	if not allows_duplicates:
		var has_impact := false
		for impact in projectile_impacts:
			has_impact = new_impact.get_script() == impact.get_script()
			if has_impact:
				return
	projectile_impacts.append(new_impact)

func get_enemy_owner() -> Enemy:
	if get_parent() is Enemy:
		return get_parent()
	
	return null

func get_fire_rate_factor() -> float:
	if player:
		return player.get_fire_rate_factor()
	else:
		return 1.0

func get_range_factor() -> float:
	if player:
		return player.get_range_factor()
	else:
		return 1.0

# Setter for emitter configuration. Whenever changed, will remove existing
# emitters and replace them with the new configuration.
func _set_emitter_configuration(value: PackedScene):
	emitter_configuration = value
	if not is_inside_tree():
		await self.ready
	
	_clear_emitters()
	_add_new_emitters()


# Setter for emitter scene. When changed, will replace existing emitters.
func _set_emitter(value: PackedScene):
	projectile_emitter = value
	_set_emitter_configuration(emitter_configuration)


# Removes existing projectile emitters.
func _clear_emitters():
	for child in get_children():
		if child is ProjectileEmitter:
			remove_child(child)
			child.queue_free()


# Adds new projectile emitters that are positioned and rotated according to the
# provided configuration scene.
func _add_new_emitters():
	var configuration := emitter_configuration.instantiate()
	
	for weapon_position in configuration.get_children():
		var new_emitter := projectile_emitter.instantiate() as ProjectileEmitter
		new_emitter.position = weapon_position.position
		new_emitter.rotation = weapon_position.rotation
		new_emitter.weapon = self
		
		add_child(new_emitter)
	
	configuration.free()

func get_volume_db_adjustment() -> float:
	# We reduce the volume depending of the number of emitters (we use cube roots to reduce less when plenty of emitters, otherwise we don't hear them)
	return additionnal_volume_db_adjustment + (pow(get_child_count() - 1, 1.0/3.0) * -4)
