extends ProjectileEmitter


@export var collisions_per_second := 4
@export var laser_width := 10

var _is_firing := false
var _current_lifetime := 0.0
var _warning_in_progress := false

@onready var _tracer := $LaserTracer
@onready var _laser_line := $Line2D as Line2D
@onready var _timer := $Timer as Timer
@onready var _casting_particles := $CastingParticles as CPUParticles2D

func _ready():
	var color: Color
	if weapon.player:
		color = Color.LIGHT_YELLOW
	else:
		color = Color.RED
	
	$Line2D.modulate = color
	$CastingParticles.modulate = color
	$LaserTracer/ImpactParticles.modulate = color

func _physics_process(_delta: float):
	_laser_line.width = max(0, (1 - (_current_lifetime/0.2)) * laser_width)
	_current_lifetime += _delta
	
	if _is_firing:
		_laser_line.points = _tracer.trace_path()


# Resets the persistent tracer's properties and position. Also keeps the direction
# vector up to date with the emitter's rotation.
func _do_fire(_delta: float, _direction: Vector2) -> bool:
	if _warning_in_progress:
		return false
	
	if not _is_firing:
		if not _warning_in_progress and not weapon.player:
			_warning_in_progress = true
			
			var warning_width := 20.0
			if weapon.projectile_motions.size() > 1: # It means we have added SineMotion (this is the only motion we can add to laser)
				warning_width = 60.0
				
			EventBus.emit_create_effect(LineWarningEffect.new(weapon.get_enemy_owner(), _direction.rotated(rotation), warning_width), global_position, false)
			await get_tree().create_timer(0.9, false).timeout
			_warning_in_progress = false
		
		_is_firing = true
		_casting_particles.emitting = true
	
	_casting_particles.rotation_degrees = rad_to_deg(_direction.angle()) + 90
	_tracer.setup(weapon, global_position, _direction.rotated(global_rotation), _get_real_max_distance())
	
	_current_lifetime = 0.0
	
	return true

func _stop_shooting():
	if _is_firing:
		_is_firing = false
		_tracer.stop()
		_casting_particles.emitting = false

func _get_real_projectiles_per_second() -> float:
	return super() * 1.2 # laser bonus

# Triggers the damage signal only when the DPS timer is not running.
func _on_projectile_collided(target: Node, _hit_location: Vector2):
	if _timer.is_stopped():
		Sounds.play_laser($Sound, weapon.get_volume_db_adjustment())
		super._on_projectile_collided(target, _hit_location)
		_timer.start(1.0 / _get_real_projectiles_per_second())

func _on_projectile_missed(_miss_location: Vector2):
	if _timer.is_stopped():
		Sounds.play_laser($Sound, weapon.get_volume_db_adjustment())
		super._on_projectile_missed(_miss_location)
		_timer.start(1.0 / _get_real_projectiles_per_second())
