# Base class for a projectile. The life cycle of the projectile follows these steps:
# 
# 1. An emitter spawns a projectile
# 2. `setup()` is called by the emitter
# 3. `_post_setup()` is called
# 4. `_update_movement()` moves the projectile according to its emitter's
# projectile motions.
# 5. The projectile eithers runs out of lifetime or hits an object.
# 
# When the projectile runs out of life, it should call `_miss()` and emit the
# `missed` signal. When it hits an object, it should call `_impact()` and emit
# the `collided` signal.
# `_miss()` and `_impact()` are there for the projectile to clean up. The signals
# are a way to communicate back to its emitter/the weapons system.
# @tags - abstract
class_name Projectile
extends CharacterBody2D


signal collided(target: Node, hit_location: Vector2)
signal missed(miss_location: Vector2)


var direction := Vector2.UP
var boomerang_phase := 0
var weapon: ModularWeapon
var volume_db_adjustments := 0.0
var max_distance := 300.0
var max_time := 5.0
var automatically_apply_rotation := true

var elapsed_time := 0.0

var _initial_global_pos := Vector2.ZERO

# Base setup function to configure spawned projectiles.
func setup(_weapon: ModularWeapon, _global_position: Vector2, _direction: Vector2, _max_distance: float):
	_initial_global_pos = _global_position
	global_position = _global_position
	max_distance = _max_distance
	direction = _direction
	weapon = _weapon
	elapsed_time = 0.0
	boomerang_phase = 0
	
	if not _weapon.player:
		if get_collision_layer_value(5):
			call_deferred("set_collision_layer_value", 5, false)
			call_deferred("set_collision_layer_value", 3, true)
		if get_collision_mask_value(4):
			call_deferred("set_collision_mask_value", 4, false)
			call_deferred("set_collision_mask_value", 2, true)
	
	_post_setup()

func _physics_process(delta: float):
	if not is_instance_valid(weapon):
		_destroy()
		return
	
	var movement := _update_movement(delta)
	if automatically_apply_rotation:
		rotation = movement.angle()
	
	var collision := move_and_collide(movement)
	
	if collision:
		collided.emit(collision.get_collider(), global_position)
		_impact()
	elif elapsed_time > max_time or _initial_global_pos.distance_to(global_position) >= max_distance:
		missed.emit(global_position)
		_miss()

# Virtual function for any specialized setup required by a derived class.
func _post_setup():
	pass

# Virtual function called when a projectile hits a target. It's where it can
# clean itself up if needed (I.E. queue_free)
func _impact():
	_destroy()

# Virtual function called when a projectile hits no target. It's where it can
# clean itself up if needed (I.E. queue_free)
func _miss():
	_destroy()

# Virtual function called when a projectile parent is destroyed. It's where it can
# clean itself up if needed (I.E. queue_free)
func _destroy():
	_stop_physics()
	
	var tween = get_tree().create_tween().bind_node(self)
	tween.tween_property(self, "scale", Vector2.ZERO, 0.1)
	tween.tween_callback(self.queue_free)

func _stop_physics():
	set_physics_process(false)
	
	call_deferred("set_collision_layer_value", 5, false)
	call_deferred("set_collision_layer_value", 3, false)

# Calculates and returns the projectile's movement this frame.
# Can mutate the projectile's state.
func _update_movement(delta: float) -> Vector2:
	elapsed_time += delta
	
	var movement_vector := Vector2.ZERO
	
	if is_instance_valid(weapon):
		if weapon.projectile_motions.is_empty():
			return movement_vector
		
		for motion in weapon.projectile_motions:
			movement_vector = motion.update_movement(self, movement_vector, direction, delta)
	else:
		_miss() # no weapon means the enemy died, let's destroy its projectiles
	
	return movement_vector

func get_distance_progression() -> float:
	return _initial_global_pos.distance_to(global_position) / max_distance
