# The base class for a projectile emitter. Its job is to spawn or control a
# projectile in its `_do_fire()` function. To actually fire a projectile, the
# base `fire()` function is provided and called, I.E. from `_unhandled_input()`.
# @tags - abstract
class_name ProjectileEmitter
extends Node2D

@export var damage_per_collision := 100

var weapon: ModularWeapon

# Base function to shoot a projectile. Calls the virtual `_do_fire` function.
func shoot(_delta: float, shooting_direction: Vector2):
	@warning_ignore("redundant_await")
	var did_fire := await _do_fire(_delta, shooting_direction)
	
	if did_fire and _delta == 0: # it means we programmatically shoot and need to call stop_shooting at one get_random_position
		await get_tree().create_timer(0.1, false).timeout
		_stop_shooting()

func stop_shooting():
	_stop_shooting()

# Virtual function that is called by `fire()`.
func _do_fire(_delta: float, _direction: Vector2) -> bool:
	return false

func _stop_shooting():
	pass

# Emits the damaged signal and triggers any impact event in the system.
func _on_projectile_collided(target: Node, hit_location: Vector2):
	weapon.damaged.emit(target, damage_per_collision)
	for impact in weapon.projectile_impacts:
		(impact as ProjectileImpact).trigger(hit_location, weapon, false, target)


# Triggers any missed event in the system.
func _on_projectile_missed(miss_location: Vector2):
	for impact in weapon.projectile_impacts:
		impact.trigger(miss_location, weapon, true, null)

func _get_random_position(allow_player_pos = true) -> Vector2:
	return find_parent("Battlefield").get_random_position(allow_player_pos)

func _get_real_projectiles_per_second() -> float:
	return weapon.projectiles_per_second * weapon.get_fire_rate_factor()

func _get_real_max_distance() -> float:
	return weapon.max_distance * weapon.get_range_factor()
