class_name HomingMotion
extends ProjectileMotion

@export var homing_radius := 60.0
@export_range(0.0, 0.5, 0.025) var homing_strength : float = 0.067

var _circle: CircleShape2D
var _space: PhysicsDirectSpaceState2D
var _query: PhysicsShapeQueryParameters2D


# Sets up a circle with the specified radius, and gets the world 2D space state.
# As we may be dealing with a projectile that checks for collisions multiple
# times per frame, we can't rely on Area2D.
func _setup_shape(_projectile: Projectile):
	_circle = CircleShape2D.new()
	_circle.radius = homing_radius
	
	_space = _projectile.get_world_2d().direct_space_state
	_query = PhysicsShapeQueryParameters2D.new()
	_query.shape = _circle
	
	if not _projectile.weapon.player:
		_query.collision_mask = 2
	else:
		_query.collision_mask = 8

# Locates the nearest target using an intersect shape call to the 2D space state.
func _find_target(_projectile: Projectile) -> Node2D:
	_query.transform = _projectile.global_transform
	var intersections := _space.intersect_shape(_query, 2)
	
	if intersections.size() > 0:
		var min_distance: float = INF
		var min_target: Node2D
		
		for intersection in intersections:
			if intersection.collider is Enemy or intersection.collider is Player: # Avoid ShieldAssist
				var distance: float = intersection.collider.global_position.distance_to(_projectile.global_position)
				if distance < min_distance:
					min_distance = distance
					min_target = intersection.collider
			
		return min_target
	
	return null

# Does not update the projectile, but updates its directions to face towards the
# target.
func update_movement(_projectile: Projectile, _movement_vector: Vector2, _direction: Vector2, _delta: float) -> Vector2:
	if not _query:
		_setup_shape(_projectile)
	
	var target := _find_target(_projectile)
	if target:
		var intended_direction := (target.global_position - _projectile.global_position).normalized()
		_projectile.direction = (_direction + intended_direction * homing_strength).normalized()
	
	return _movement_vector
